import asyncio
import sys
import tomllib
from collections.abc import Sequence
from dataclasses import dataclass
from enum import StrEnum
from pathlib import Path

from .formatting import bright, dim, status
from .utils import exec_command, human_readable_duration, timed_awaitable


class BoredomError(Exception):
    pass


@dataclass
class ConfigError(Exception):
    config_file: Path
    message: str


class Output(StrEnum):
    OUT = 'info'
    ERR = 'error'


class Verbosity:
    def __init__(self, success: Sequence[Output] = None, fail: Sequence[Output] = None):
        self.success = success or []
        self.fail = fail or [Output.OUT, Output.ERR]

    def should_output(self, out: Output, success: bool):
        return (out in self.success) if success else (out in self.fail)


class TaskRunner:
    def __init__(self, tasks: dict[str, str], verbosity: Verbosity):
        if not tasks:
            raise BoredomError
        self.tasks = tasks
        self.verbosity = verbosity

    @classmethod
    def with_pyproject_config(cls) -> 'TaskRunner':
        pyproject_file = Path('pyproject.toml')
        try:
            with pyproject_file.open('rb') as f:
                config = tomllib.load(f).get('tool', {}).get('powerchord', {})
        except OSError as exc:
            raise ConfigError(pyproject_file, str(exc)) from exc
        if not config:
            raise ConfigError(pyproject_file, 'Could not find any [tool.powerchord(.*)] entries')
        return cls(config.get('tasks', {}), Verbosity(**config.get('verbosity', {})))

    async def run_task(self, name: str, task: str) -> tuple[str, bool]:
        (success, out, err), duration = await timed_awaitable(
            exec_command(task),
            formatter=lambda t: dim(f'[{human_readable_duration(t)}]'),
        )
        sys.stdout.write(f'{status(success)} {name} {duration}\n')
        if self.verbosity.should_output(Output.OUT, success):
            sys.stdout.buffer.write(out)
            sys.stdout.buffer.flush()
        if self.verbosity.should_output(Output.ERR, success):
            sys.stderr.buffer.write(err)
            sys.stderr.buffer.flush()
        return name, success

    async def run_tasks(self) -> list[tuple[str, bool]]:
        sys.stdout.write(bright('To do:\n'))
        for name, task in self.tasks.items():
            sys.stdout.write(f'• {name}: {dim(f"[{task}]")}\n')
        sys.stdout.write(bright('\nResults:\n'))
        futures = [
            asyncio.create_task(self.run_task(name, task))
            for name, task in self.tasks.items()
        ]
        return [await f for f in futures]


def fail_with(*lines: str) -> None:
    sys.exit('💀 ' + '\n'.join(lines))


def run_tasks() -> None:
    try:
        task_runner = TaskRunner.with_pyproject_config()
    except ConfigError as exc:
        fail_with(f'Error while loading {exc.config_file}:\n{exc.message}')
    except BoredomError:
        sys.stdout.write('Nothing to do. Getting bored...\n')
    else:
        failed_tasks = [name for name, success in asyncio.run(
            task_runner.run_tasks(),
        ) if not success]
        if failed_tasks:
            sys.stderr.write('\n')
            fail_with(bright('Failed tasks:'), ', '.join(failed_tasks))



from collections.abc import Callable


def _wrap_ansi_code(value: int) -> Callable[[str], str]:
    def wrapper(s: str) -> str:
        return f'\033[{value}m{s}\033[0m'
    return wrapper


bright = _wrap_ansi_code(1)
dim = _wrap_ansi_code(2)

black = _wrap_ansi_code(30)
red = _wrap_ansi_code(31)
green = _wrap_ansi_code(32)
yellow = _wrap_ansi_code(33)
blue = _wrap_ansi_code(34)
magenta = _wrap_ansi_code(35)
cyan = _wrap_ansi_code(36)
gray = _wrap_ansi_code(37)

light_gray = _wrap_ansi_code(90)
light_red = _wrap_ansi_code(91)
light_green = _wrap_ansi_code(92)
light_yellow = _wrap_ansi_code(93)
light_blue = _wrap_ansi_code(94)
light_magenta = _wrap_ansi_code(95)
light_cyan = _wrap_ansi_code(96)
white = _wrap_ansi_code(97)

OK = green('✔')
FAIL = red('✘')


def status(success: bool) -> str:
    return OK if success else FAIL


# def status(success: bool) -> Callable[[str], str]:
#     def wrapper(s: str) -> str:
#         return f'{OK if success else FAIL} {s}'
#     return wrapper
#
#
# ok = status(True)
# fail = status(False)



import asyncio
from collections.abc import Awaitable, Callable
from subprocess import PIPE
from time import perf_counter_ns
from typing import TypeVar

T = TypeVar('T')


async def exec_command(command: str) -> tuple[bool, bytes, bytes]:
    proc = await asyncio.create_subprocess_shell(command, stdout=PIPE, stderr=PIPE)
    out, err = await proc.communicate()
    return proc.returncode == 0, out, err


def human_readable_duration(nanoseconds: int) -> str:
    minutes = int(nanoseconds // 60_000_000_000)
    nanoseconds %= 60_000_000_000
    seconds = int(nanoseconds // 1_000_000_000)
    nanoseconds %= 1_000_000_000
    milliseconds = int(nanoseconds // 1_000_000)
    nanoseconds %= 1_000_000
    microseconds = int(nanoseconds // 1_000)
    nanoseconds %= 1_000
    if minutes:
        return f'{minutes:d}:{seconds:02d}.{milliseconds:03d} minutes'
    if seconds:
        return f'{seconds:d}.{milliseconds:03d} seconds'
    if milliseconds:
        return f'{milliseconds:d}.{microseconds:03d} ms'
    return f'{microseconds:d}.{nanoseconds:03d} µs'


def timed(
    func: Callable[[], T],
    formatter: Callable[[int], str] = None,
) -> tuple[T, str]:
    start = perf_counter_ns()
    return func(), (formatter or human_readable_duration)(perf_counter_ns() - start)


def timed_awaitable(
    awaitable: Awaitable[T],
    formatter: Callable[[int], str] = None,
) -> Awaitable[tuple[T, str]]:
    async def wrapper() -> tuple[T, str]:
        start = perf_counter_ns()
        return await awaitable, (formatter or human_readable_duration)(perf_counter_ns() - start)
    return wrapper()
